#include <linux/compiler.h>
#include <linux/bitops.h>
#include <linux/bootmem.h>
#include <linux/bootmem_tools.h>
#include <linux/mmzone.h>
#include <linux/spinlock.h>

void dump_bootmem_core(bootmem_data_t * bdata)
{
	char *format = "0x%08lx - 0x%08lx (%4lup): %s\n";
	char *used = "used";
	char *free = "free";
	char *state;
	unsigned long i = 0, j;
	unsigned long eidx =
	    bdata->node_low_pfn - (bdata->node_boot_start >> PAGE_SHIFT);
	unsigned long startpfn, endpfn;

#ifdef CONFIG_DEBUG_BOOTMEM
	printk("bdata->node_boot_start = %lu\n", bdata->node_boot_start);
	printk("bdata->node_low_pfn    = %lu\n", bdata->node_low_pfn);
	printk("                  eidx = %lu\n", eidx);
#endif

	while (i < eidx) {
		if (test_bit(i, bdata->node_bootmem_map)) {
			j = find_next_zero_bit(bdata->node_bootmem_map,
					       eidx, i);
			state = used;
		} else {
			j = find_next_bit(bdata->node_bootmem_map, eidx,
					  i);
			state = free;
		}
		j = min(j, eidx);

		startpfn = i + bdata->node_boot_start;
		endpfn = j + bdata->node_boot_start;
		printk(format, startpfn << PAGE_SHIFT,
		       endpfn << PAGE_SHIFT, j - i, state);

		i = j;
	}
}

void dump_bootmem(void)
{
	unsigned long flags;
	spin_lock_irqsave(&bootmem_lock, flags);
	dump_bootmem_core(NODE_DATA(0)->bdata);
	spin_unlock_irqrestore(&bootmem_lock, flags);
}

static unsigned long usable_bootmem_core(bootmem_data_t * bdata)
{
	unsigned long i;
	unsigned long eidx =
	    bdata->node_low_pfn - (bdata->node_boot_start >> PAGE_SHIFT);
	unsigned long freepages = 0;

	for (i = 0; i < eidx; ++i)
		if (!test_bit(i, bdata->node_bootmem_map))
			++freepages;

	return freepages;
}

unsigned long usable_bootmem(void)
{
	unsigned long flags, rv = 0;
	spin_lock_irqsave(&bootmem_lock, flags);
	rv = usable_bootmem_core(NODE_DATA(0)->bdata);
	spin_unlock_irqrestore(&bootmem_lock, flags);
	return rv;
}

static unsigned long _gcd(unsigned long a, unsigned long b)
{
	unsigned long c;
	while ((c = a % b)) {
		a = b;
		b = c;
	}
	return b;
}

/* gives some info about the free block:
 * - freeblocks: count of free blocks
 * - totalfree: count of total free pages
 * - maxfree: count of pages in the largest free block
 * - gcd: greatest common divisor (# pages) of the free blocks pages
 *        (single pages are ignored, one is returned as a minimum
 *         also when no free page was found)
 * should be called with bootmem_lock hold
 */
void bootmem_freeinfo(unsigned long *freeblocks, unsigned long *totalfree,
		      unsigned long *minfree, unsigned long *maxfree,
		      unsigned long *gcd)
{
	bootmem_data_t *bdata = NODE_DATA(0)->bdata;
	int is_free;
	unsigned long i = 0, j;
	unsigned long eidx =
	    bdata->node_low_pfn - (bdata->node_boot_start >> PAGE_SHIFT);
	unsigned long cfreeblocks = 0, ctotalfree = 0;
	unsigned long cminfree = 0, cmaxfree = 0, cgcd = 0;

	while (i < eidx) {
		if (test_bit(i, bdata->node_bootmem_map)) {
			j = find_next_zero_bit(bdata->node_bootmem_map,
					       eidx, i);
			is_free = 0;
		} else {
			j = find_next_bit(bdata->node_bootmem_map, eidx,
					  i);
			is_free = 1;
			++cfreeblocks;
		}
		j = min(j, eidx);

		if (is_free) {
			ctotalfree += j - i;

			if (unlikely(cminfree == 0))
				cminfree = j - i;
			cminfree = min(j - i, cminfree);

			cmaxfree = max(j - i, cmaxfree);

			if ((j - i) > 1) {
				if (unlikely(cgcd == 0))
					cgcd = j - i;
				cgcd = _gcd(j - i, cgcd);
			}
		}

		i = j;
	}

	if (freeblocks)
		*freeblocks = cfreeblocks;
	if (totalfree)
		*totalfree = ctotalfree;
	if (minfree)
		*minfree = cminfree;
	if (maxfree)
		*maxfree = cmaxfree;
	if (gcd)
		*gcd = cgcd;
}

extern void *____alloc_bootmem_core(struct bootmem_data *bdata,
				    unsigned long size,
				    unsigned long align,
				    unsigned long goal,
				    unsigned long limit);

void *alloc_maxfree_bootmem(int *size)
{
	unsigned long flags;
	unsigned long max = 0;
	void *rv = NULL;

	if (size)
		*size = 0;

	spin_lock_irqsave(&bootmem_lock, flags);
#ifdef CONFIG_DEBUG_BOOTMEM
	printk("Bootmem before: alloc_maxfree_bootmem\n");
	dump_bootmem_core(NODE_DATA(0)->bdata);
#endif
	bootmem_freeinfo(NULL, NULL, NULL, &max, NULL);

	if (!max)
		goto ret;

	rv = ____alloc_bootmem_core(NODE_DATA(0)->bdata, max << PAGE_SHIFT,
				    0, 0, 0);
	if (rv && size)
		*size = max << PAGE_SHIFT;

      ret:
#ifdef CONFIG_DEBUG_BOOTMEM
	printk("Bootmem after: alloc_maxfree_bootmem\n");
	dump_bootmem_core(NODE_DATA(0)->bdata);
#endif
	spin_unlock_irqrestore(&bootmem_lock, flags);
	return rv;
}
